@@ -3,12 +3,22 @@
  *
  * This file is part of GnuPG.
  *
- * GnuPG is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 3 of the License, or
- * (at your option) any later version.
+ * This file is free software; you can redistribute it and/or modify
+ * it under the terms of either
  *
- * GnuPG is distributed in the hope that it will be useful,
+ *   - the GNU Lesser General Public License as published by the Free
+ *     Software Foundation; either version 3 of the License, or (at
+ *     your option) any later version.
+ *
+ * or
+ *
+ *   - the GNU General Public License as published by the Free
+ *     Software Foundation; either version 2 of the License, or (at
+ *     your option) any later version.
+ *
+ * or both in parallel, as here.
+ *
+ * This file is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
@@ -55,12 +65,16 @@ enum
     STATUS_GOODMDC,
     STATUS_BADMDC,
     STATUS_ERRMDC,
+
     STATUS_IMPORTED,
     STATUS_IMPORT_OK,
     STATUS_IMPORT_PROBLEM,
     STATUS_IMPORT_RES,
     STATUS_IMPORT_CHECK,
 
+    STATUS_EXPORTED,
+    STATUS_EXPORT_RES,
+
     STATUS_FILE_START,
     STATUS_FILE_DONE,
     STATUS_FILE_ERROR,
@@ -119,12 +133,20 @@ enum
     STATUS_PKA_TRUST_BAD,
     STATUS_PKA_TRUST_GOOD,
 
+    STATUS_TOFU_USER,
+    STATUS_TOFU_STATS,
+    STATUS_TOFU_STATS_SHORT,
+    STATUS_TOFU_STATS_LONG,
+
     STATUS_TRUNCATED,
     STATUS_MOUNTPOINT,
 
     STATUS_PINENTRY_LAUNCHED,
 
+    STATUS_PLAINTEXT_FOLLOWS,   /* Used by g13-syshelp  */
+
     STATUS_ERROR,
+    STATUS_WARNING,
     STATUS_SUCCESS,
     STATUS_FAILURE,
 
